/* ********************************************************************* *
   tag: mod_auth_tkt
   MD5 based ticket-type user authentication with optional timeout and 
   authentication of user supplied data 
 * ********************************************************************* */

/* =======================================================================
 * Portions Copyright (c) 2001-2002 Open Fusion Pty. Ltd. Australia.
 * All rights reserved. 
 * Portions Copyright (c) 2000 Liquid Digital Information Systems, Inc. 
 * All rights reserved. 
 *
 * This software was written for Open Fusion Pty. Ltd. (Australia) by 
 * Gavin Carr, and for Liquid Digital Information Systems, Inc. 
 * by Raimondas Kiveris. It is hereby contributed to the Apache Software 
 * Foundation for distribution under the Apache license, as follows.
 *	
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by the Apache Group
 *  for use in the Apache HTTP server project (http://www.apache.org/)."
 * 
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *  endorse or promote products derived from this software without
 *  prior written permission. For written permission, please contact
 *  apache@apache.org.
 * 
 * 5. Products derived from this software may not be called "Apache"
 *  nor may "Apache" appear in their names without prior written
 *  permission of the Apache Group.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *  acknowledgment:
 *  "This product includes software developed by the Apache Group
 *  for use in the Apache HTTP server project (http://www.apache.org/)."
 * 
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * =======================================================================
 *	
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 * 
 * ======================================================================
 */


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include "tkt_digest.h"

#include "test_char.h"
#define TEST_CHAR(c, f) (test_char_table[(unsigned)(c)] & (f))
static const char c2x_table[] = "0123456789abcdef";

#include "mod_auth_tkt.h"

module MODULE_VAR_EXPORT tkt_auth_module;


/*****************************************************************/

#ifdef SOLARIS2
/**
 * strsep() is not part of the Solaris libraries
 * 
 * strsep - Split a string into tokens
 * @s: The string to be searched
 * @ct: The characters to search for
 *
 * strsep() updates @s to point after the token, ready for the next call.
 *
 * It returns empty tokens, too, behaving exactly like the libc function
 * of that name. In fact, it was stolen from glibc2 and de-fancy-fied.
 * Same semantics, slimmer shape. ;)
 *
 * By Christian Folinic <folinic@post.ch>
 */

char * strsep(char **s, const char *ct)
{
	char *sbegin = *s, *end;

	if (sbegin == NULL)
		return NULL;

	end = strpbrk(sbegin, ct);
	if (end)
		*end++ = '\0';
	*s = end;

	return sbegin;
}
#endif

/*****************************************************************/

/* initializer */

void tkt_auth_init (server_rec *s, pool *p)
{
    ap_add_version_component("mod_auth_tkt/" TKT_AUTH_VERSION);
}


/* dir config creator */

static void * 
create_tkt_auth_config(pool *p, char* path)
{
  tkt_auth_dir_conf *new =
    (tkt_auth_dir_conf *) ap_palloc(p, sizeof(tkt_auth_dir_conf));

  new->directory = path;
  new->login_url = NULL;
  new->timeout_url = NULL;
  new->unauth_url = NULL;
  new->auth_domain = NULL;
  new->auth_token = ap_make_array ( p, 0, sizeof (char *) );
  new->auth_cookie_name = NULL;
  new->back_cookie_name = NULL;
  new->back_arg_name = NULL;
  new->ignore_ip = 0;
  new->timeout_sec = 0;
  new->timeout_refresh = .5;

  return (void *) new;  
}

/* -------------------------------------- */



/* dir merger */

static void * 
merge_tkt_auth_config(pool *p, void* parent_dirv, void* subdirv)
{
  tkt_auth_dir_conf *parent = (tkt_auth_dir_conf *) parent_dirv;
  tkt_auth_dir_conf *subdir = (tkt_auth_dir_conf *) subdirv;
  tkt_auth_dir_conf *new =
    (tkt_auth_dir_conf *) ap_palloc(p, sizeof(tkt_auth_dir_conf));

  new->directory = (subdir->directory) ? subdir->directory : parent->directory;
  new->login_url = (subdir->login_url) ? subdir->login_url : parent->login_url;
  new->timeout_url = (subdir->timeout_url) ? subdir->timeout_url : parent->timeout_url;
  new->unauth_url = (subdir->unauth_url) ? subdir->unauth_url : parent->unauth_url;
  new->auth_domain = (subdir->auth_domain) ? subdir->auth_domain : parent->auth_domain;
  new->auth_token = (subdir->auth_token->nelts > 0) ? subdir->auth_token : parent->auth_token;
  new->auth_cookie_name = (subdir->auth_cookie_name) ? subdir->auth_cookie_name : parent->auth_cookie_name;
  new->back_cookie_name = (subdir->back_cookie_name) ? subdir->back_cookie_name : parent->back_cookie_name;
  new->back_arg_name = (subdir->back_arg_name) ? subdir->back_arg_name : parent->back_arg_name;
  new->ignore_ip = (subdir->ignore_ip) ? subdir->ignore_ip : parent->ignore_ip;
  new->timeout_sec = (subdir->timeout_sec) ? subdir->timeout_sec : parent->timeout_sec;
  new->timeout_refresh = (subdir->timeout_refresh) ? subdir->timeout_refresh : parent->timeout_refresh;

  return (void *) new;
}

/* -------------------------------------- */



/* server config */

static void *
create_tkt_auth_serv_config(pool *p, server_rec* s)
{
  tkt_auth_serv_conf *new =
    (tkt_auth_serv_conf *) ap_palloc(p, sizeof(tkt_auth_serv_conf));
  new->md5_key = NULL;

  return (void *) new;
} 

/* -------------------------------------- */



/* merge server config */

/* -------------------------------------- */



/* command table */

static const command_rec tkt_auth_cmds[] =
{
  {"TKTAuthLoginURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, login_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if authentication fails."},
  {"TKTAuthTimeoutURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, timeout_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if cookie times-out."},
  {"TKTAuthUnauthURL", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, unauth_url),
   OR_AUTHCFG, TAKE1, "where to redirect to if valid user without required token."},
  {"TKTAuthIgnoreIP", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, ignore_ip),
   OR_AUTHCFG, TAKE1, "ip address to use instead of remote addr."},
  {"TKTAuthCookieName", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_cookie_name),
   OR_AUTHCFG, TAKE1, "name to use for ticket cookie."},
  {"TKTAuthBackCookieName", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, back_cookie_name),
   OR_AUTHCFG, TAKE1, "name to use for back cookie (null for none)."},
  {"TKTAuthBackArgName", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, back_arg_name),
   OR_AUTHCFG, TAKE1, "name to use for back url argument (null for none)."},
  {"TKTAuthDomain", ap_set_string_slot, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_domain),
   OR_AUTHCFG, TAKE1, "domain to use in cookies."},
  {"TKTAuthToken", set_tkt_auth_token, (void *) XtOffsetOf(tkt_auth_dir_conf, auth_token),
   OR_AUTHCFG, ITERATE, "token required to access this area (NULL for none)."},
  {"TKTAuthTimeoutMin", set_tkt_auth_timeout, 
   NULL, OR_AUTHCFG, TAKE1, "cookie inactivity timeout in minutes."},
  {"TKTAuthTimeoutRefresh", set_tkt_auth_timeout_refresh, 
   NULL, OR_AUTHCFG, TAKE1, "cookie timeout refresh flag."},
  {"TKTAuthSecret", setup_md5key, NULL, 
   RSRC_CONF, TAKE1, "secret key to use in MD5 digest."},

  {NULL},
};

/* -------------------------------------- */


/* command specific functions */

static const char *
set_tkt_auth_token ( cmd_parms *cmd, void *cfg, char *param )
{
  char	**new;
  tkt_auth_dir_conf *dir_cfg = (tkt_auth_dir_conf *) cfg;

  new = (char **) ap_push_array ( dir_cfg->auth_token );
  *new = ap_pstrdup ( cmd->pool, param );

  return NULL;
}


static const char *
setup_md5key (cmd_parms *cmd, void *dir_cfg, char *param)
{
  tkt_auth_serv_conf *serv_cfg = (tkt_auth_serv_conf *)
    ap_get_module_config(cmd->server->module_config, &tkt_auth_module);

  serv_cfg->md5_key = param;
  return NULL;
}

static const char *
set_tkt_auth_timeout(cmd_parms *cmd, void *cfg, char *param)
{
  tkt_auth_dir_conf *dir_cfg = (tkt_auth_dir_conf *)cfg;
  
  long int minutes = strtol (param, (char **) NULL, 10);
  
  if (minutes < 0) 
    return ("Timeout must be positive");
  
  if (minutes == LONG_MAX) 
    return ("Integer overflow or invalid number");
  
  dir_cfg->timeout_sec = minutes * 60;
  
  return (NULL);
}

static const char *
set_tkt_auth_timeout_refresh(cmd_parms *cmd, void *cfg, char *param)
{
  tkt_auth_dir_conf *dir_cfg = (tkt_auth_dir_conf *)cfg;
  
  double refresh = atof(param);

  if (refresh < 0 || refresh > 1) 
    return ("Refresh flag must be between 0 and 1");
  
  dir_cfg->timeout_refresh = refresh;
  
  return (NULL);
}

/* -------------------------------------- */


/* handlers */
/* -------------------------------------- */



/* filename translation */
/* -------------------------------------- */



/* check_user_id */

static int 
authenticate_tkt_user (request_rec *r)
{
  tkt_auth_dir_conf *tkt_cfg = 
    ap_get_module_config(r->per_dir_config, &tkt_auth_module);
  tkt_auth_serv_conf *tkt_scfg =
    ap_get_module_config(r->server->module_config, &tkt_auth_module);

  char *cookie = NULL;
  char *url_cookie = NULL;
  tkt_cookie * parsed;
  int rc;
  unsigned int now = time(NULL);
  char *digest;
  char *remote_ip; 
  char *domain;
  int i;
  
  /* Default cookie name */
  if (! tkt_cfg->auth_cookie_name) {
    tkt_cfg->auth_cookie_name = AUTH_COOKIE_NAME;
  }
  /* Default back arg name */
  if (! tkt_cfg->back_cookie_name && ! tkt_cfg->back_arg_name) {
    tkt_cfg->back_arg_name = BACK_ARG_NAME;
  }

  /* Module not configured unless login_url is set */
  if (! tkt_cfg->login_url) {	
    return DECLINED;
  }

  /* Setup cookie domain - default to current domain if auth_domain not set */
  domain = tkt_cfg->auth_domain;
  if (! domain) 
    domain = strchr(ap_get_server_name(r), '.');

#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
    "TKT: tkt_cfg: (is_main: %d, is_first: %d) login_url %s, auth_cookie_name %s, auth_domain %s, # auth_tokens %d, timeout_sec %lu, uri %s", 
    (r->main == NULL), 
    (r->prev == NULL),
    tkt_cfg->login_url, 
    tkt_cfg->auth_cookie_name,
    tkt_cfg->auth_domain,
    tkt_cfg->auth_token->nelts,
    tkt_cfg->timeout_sec,
    r->unparsed_uri
  );
#endif

  if (! tkt_scfg->md5_key) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "TKT: TKTAuthSecret missing");
    return SERVER_ERROR;
  }

  remote_ip = (tkt_cfg->ignore_ip != NULL && strncmp(tkt_cfg->ignore_ip,"0",1) != 0) ? "0.0.0.0" : r->connection->remote_ip;

  parsed = ap_palloc(r->pool, sizeof(tkt_cookie));

  /* Check for valid ticket in url */
  url_cookie = find_our_url(r, tkt_cfg);

  if (url_cookie != NULL) {
    /* Treat empty url cookie as reset, for logouts */
    if (strlen(url_cookie) == 0) {
      char *cookie;
#ifdef DEBUG
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: empty url ticket found, redirecting to %s", tkt_cfg->login_url);
#endif
      
      /* Reset any auth cookie we have */
      cookie = domain ?
	ap_psprintf(r->pool, 
          "%s=; path=/; domain=%s", 
          tkt_cfg->auth_cookie_name,
          domain) :
	ap_psprintf(r->pool, 
          "%s=; path=/",
          tkt_cfg->auth_cookie_name);

      ap_table_setn(r->err_headers_out, "Set-Cookie", cookie);
    
      return (redirect(r, tkt_cfg->login_url, tkt_cfg));
    }

    /* url cookie found, test validity */
    else {
      rc = parse_cookie(r, &url_cookie, parsed);
      if (rc == 0) {
	cookie = url_cookie;
#ifdef DEBUG
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
	  "TKT: valid url ticket found (%s)", url_cookie);
#endif
      }
    }
  }

  /* If no url ticket, check auth cookie */
  if (cookie == NULL) {
    cookie = find_our_cookie(r, tkt_cfg);
    if (cookie != NULL && strlen(cookie) > 0) {
      rc = parse_cookie(r, &cookie, parsed);
      /* Invalid cookie */
      if (rc != 0) {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, 
          "TKT: invalid cookie ticket found (%s)", cookie);
        return (redirect(r, tkt_cfg->login_url, tkt_cfg));
      }
#ifdef DEBUG
      else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: valid cookie ticket found (%s)", cookie);
      }
#endif
    }
    /* Missing cookie */
    else {
      if (url_cookie != NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, 
          "TKT: invalid url ticket found (%s)", url_cookie);
      }
#ifdef DEBUG
      else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: no ticket found in cookie or url, redirecting to %s", 
            tkt_cfg->login_url);
      }
#endif
    
      return (redirect(r, tkt_cfg->login_url, tkt_cfg));
    }
  }

#ifdef DEBUG
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
      "TKT: parsed cookie: uid %s, timestamp %d, tokens '%s', data '%s'",
      parsed->uid, parsed->timestamp, parsed->tokens, parsed->user_data);
#endif

  /* Check timeout */
  if (tkt_cfg->timeout_sec > 0) {
    
    if (now > tkt_cfg->timeout_sec + parsed->timestamp) {
      char *timeout_date, *timeout_cookie;
#ifdef DEBUG
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: cookie timed out: now %d, timeout: %ld, tstamp: %d", 
        now, tkt_cfg->timeout_sec, parsed->timestamp);
#endif
      
      /* Delete cookie if invalid, in case we want to set from the url */
      timeout_date = ap_ht_time(r->pool, time(NULL) - 3600, "%a %d %b %Y %T %Z", 0);
      timeout_cookie = domain ?
	ap_psprintf(r->pool,
          "%s=; path=/; domain=%s; expires=%s",
          tkt_cfg->auth_cookie_name, domain, timeout_date) :
	ap_psprintf(r->pool,
          "%s=; path=/; expires=%s",
          tkt_cfg->auth_cookie_name, timeout_date);

      ap_table_setn(r->err_headers_out, "Set-Cookie", timeout_cookie);                                                                                      
      return redirect(r, tkt_cfg->timeout_url ? 
                      tkt_cfg->timeout_url : tkt_cfg->login_url,
                      tkt_cfg);
    } 
#ifdef DEBUG
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: cookie timeout still good: now %d, timeout: %ld, tstamp: %d", 
        now, tkt_cfg->timeout_sec, parsed->timestamp);
    }
#endif
  }

  digest = tkt_digest(remote_ip,        /* digest is malloc'ed in tkt_digest */
		      tkt_scfg->md5_key, 
		      parsed->uid, 
                      parsed->tokens,
		      parsed->user_data, 
		      parsed->timestamp);

#ifdef DEBUG
  /*
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
      "TKT: digest check on cookie %-32.32s, digest %s, "
      "ip %s, key %s, uid %s, timestamp %d, tokens '%s', data '%s'",
      cookie, digest,
      remote_ip, tkt_scfg->md5_key, parsed->uid, 
      parsed->timestamp, parsed->tokens, parsed->user_data);
  */
#endif

  if (!memcmp(cookie, digest, MD5_DIGEST_SZ)) { /* user is authenticated */
    char buf[640];
    char *cookie;
    
    free (digest);
    
    /* check for auth_token, if required */
    if (tkt_cfg->auth_token->nelts > 0 && 
        strcmp(((char **) tkt_cfg->auth_token->elts)[0], "NULL") != 0) {
      int unauth = 1;
      char buf[640];
      char *b = buf;
      char *next_parsed_token;

      snprintf(buf, sizeof buf, "%s", parsed->tokens);

      /* Check tokens - any token present in both lists is a match */
      while ((next_parsed_token = strsep(&b, ","))) {
        int i;
        char ** auth_tokens = (char **) tkt_cfg->auth_token->elts;

        for (i=0 ; i<tkt_cfg->auth_token->nelts; ++i) {
          int token_len = strlen(auth_tokens[i]);
          if (!strncmp(auth_tokens[i], next_parsed_token, token_len)) {
            unauth = 0;
            goto out;
          }
        }
      }
      out:

      /* Redirect if not authorised */
      if (unauth) {
#ifdef DEBUG
        /*
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: no matching token found, redirecting");   
	*/
#endif
        return (redirect(r, 
	    tkt_cfg->unauth_url ?  tkt_cfg->unauth_url : tkt_cfg->login_url, 
	    tkt_cfg));
      }
#ifdef DEBUG
      else {
	/*
	ap_log_rerror ( APLOG_MARK, APLOG_DEBUG, r,
		"TKT: matching token found: '%s'", next_parsed_token );
        */
      }
#endif
    }


    /* Refresh the cookie if using inactivity timeouts */
    if (tkt_cfg->timeout_sec > 0 && tkt_cfg->timeout_refresh > 0) {

      /* The timeout refresh is a double between 0 and 1, signifying what
       * proportion of the timeout should be left before we refresh i.e. 
       * 0 means never refresh (hard timeouts); 1 means always refresh;
       * .33 means only refresh after two-thirds of the timeout period has 
       * elapsed. */ 
      int elapsed = now - parsed->timestamp;
      double refresh_sec = (1 - tkt_cfg->timeout_refresh) * tkt_cfg->timeout_sec;
#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: timeout %d, refresh %.3f, refresh_sec %.3f, elapsed %d",
	    tkt_cfg->timeout_sec, tkt_cfg->timeout_refresh, refresh_sec, elapsed);
#endif
      if (elapsed > refresh_sec) {
      
        /* Reconstruct the cookie  */
        digest = tkt_digest(remote_ip,
			    tkt_scfg->md5_key, 
			    parsed->uid, 
			    parsed->tokens, 
			    parsed->user_data, 
			    now);
      
        memcpy (buf, digest, 32);
        if (parsed->tokens) {
            sprintf(&buf[32], "%08x%s%c%s%c%s", 
                    now, 
                    parsed->uid, 
                    SEPARATOR,
                    parsed->tokens,
                    SEPARATOR,
                    parsed->user_data);
        }
        else {
            sprintf(&buf[32], "%08x%s%c%s", 
                    now, 
                    parsed->uid, 
                    SEPARATOR,
                    parsed->user_data);
        }
        
        {
	  char *copy = strdup (buf);
	  ap_base64encode(buf, copy, strlen(copy));
	  free(copy);
        }
      
#ifdef DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
            "TKT: refreshing cookie: %s=%s, domain=%s",
	        tkt_cfg->auth_cookie_name, buf, domain);
#endif

        cookie = domain ? 
	  ap_psprintf(r->pool, 
            "%s=%s; path=/; domain=%s", 
	    tkt_cfg->auth_cookie_name, buf, domain) :
	  ap_psprintf(r->pool, 
            "%s=%s; path=/", 
	    tkt_cfg->auth_cookie_name, buf);
      
        ap_table_setn (r->err_headers_out, "Set-Cookie", cookie);
        free(digest);
      }
    }
    
    r->connection->user = parsed->uid;
    r->connection->ap_auth_type = "Basic";	/* for other modules */
    
    ap_table_set(r->subprocess_env, REMOTE_USER_DATA_ENV, parsed->user_data);
    ap_table_set(r->subprocess_env, REMOTE_USER_TOKENS_ENV, parsed->tokens);

#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
        "TKT: user '%s' authenticated, tokens '%s', data '%s'", 
            parsed->uid, parsed->tokens, parsed->user_data);
#endif
 
    return (OK);
  }

  ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, 
        "TKT: someone tampered with cookie <%s>", cookie);
  
  free(digest);
  return (redirect(r, tkt_cfg->login_url, tkt_cfg));
}

/* -------------------------------------- */


/* check auth */
/* -------------------------------------- */


/* check access */
/* -------------------------------------- */


/* type_checker */
/* -------------------------------------- */


/* fixups */
/* -------------------------------------- */


/* logger */
/* -------------------------------------- */


/* header parser */
/* -------------------------------------- */


/* chitkt_init */
/* -------------------------------------- */


/* chitkt_exit */
/* -------------------------------------- */


/* post read-request */
/* -------------------------------------- */


/* misc functions */

static int 
parse_cookie(request_rec *r, char **magic, tkt_cookie *parsed)
{
  char *cookie = *magic;
  int len = strlen (cookie);
  int sepidx, sep2idx;
  
  if (len <= MD5andTSTAMP) {
    return (1); /* not our cookie */
  }
  
  /* see if there is a uid/data separator */
  sepidx = ap_ind(cookie, SEPARATOR);
  
  if (sepidx == -1) {	/* cookie either in Base64 format, 
			 * has URI hex escapes or is corrupted */
    if (strstr(cookie, SEPARATOR_HEX) ) { /* hex('!')  */
      ap_unescape_url(cookie);
      
      sepidx = ap_ind(cookie, SEPARATOR);
    }
    else {
      char *buf = (char *) ap_palloc(r->pool, len);  /* Base64 encoded string is 
						      * longer than original */
      ap_base64decode(buf, cookie);
      
      sepidx = ap_ind (buf, SEPARATOR);
      
      if (sepidx == -1) {
	return (1);				  /* not our cookie */
      }
      
      cookie = *magic = buf;
    }
    
    len = strlen(cookie);
  }

#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie decoded cookie: %s", cookie); */
#endif
  
  /* get the user id */
  parsed->uid = ap_palloc(r->pool, sepidx - MD5andTSTAMP + 1);
  memcpy(parsed->uid, &cookie[MD5andTSTAMP], sepidx - MD5andTSTAMP);
  parsed->uid[sepidx - MD5andTSTAMP] = '\0';
  
  /* check for tokens */
  sep2idx = ap_ind (&cookie[sepidx+1], SEPARATOR);
  if (sep2idx == -1) {
#ifdef DEBUG
      /*
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie: no tokens");
      */
#endif
      parsed->tokens = ap_palloc(r->pool, 1);
      *parsed->tokens = '\0';
  }
  else {
      /* swap sepidx and sep2idx */
      int tmp = sepidx;
      sepidx = tmp + sep2idx + 1;
      sep2idx = tmp;
#ifdef DEBUG
/*    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie: tokens found - sep2=%d, sep=%d, len=%d", sep2idx, sepidx, len); */
#endif
      parsed->tokens = ap_palloc(r->pool, sepidx - sep2idx);
      snprintf (parsed->tokens, sepidx-sep2idx, "%s", &cookie[sep2idx+1]);
#ifdef DEBUG
/*    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT parse_cookie tokens: %s", parsed->tokens); */
#endif
  }

  /* user data sent in a cookie */
  parsed->user_data = ap_palloc(r->pool, len - sepidx);
  snprintf (parsed->user_data, len-sepidx, "%s", &cookie[sepidx+1]);
  
  /* and timestamp */
  sscanf(&cookie[MD5_DIGEST_SZ], "%8x", &(parsed->timestamp));
  
  return (0);
}


static int 
cookie_match(void *result, const char *key, const char *cookie)
{
  char *value;
  cookie_res * cr = (cookie_res *) result;
  
  if (cookie != NULL) {
#ifdef DEBUG
    /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match, key %s against <%s> (name=%s)", key, cookie, cr->cookie_name); */
#endif

    char *cookie_name = ap_palloc(cr->r->pool, strlen(cr->cookie_name) + 2);
    strncpy(cookie_name, cr->cookie_name, strlen(cr->cookie_name));
    cookie_name[strlen(cr->cookie_name)] = '=';
    cookie_name[strlen(cr->cookie_name) + 1] = '\0';
    if ((value = strstr(cookie, cookie_name))) {
      char *cookiebuf, *cookieend;
      
      value += strlen(cookie_name);
      cookiebuf = ap_pstrdup(cr->r->pool, value);
      cookieend = strchr(cookiebuf, ';');
      if (cookieend) {
	*cookieend = '\0';      /* Ignore anything after the next ; */
      }
      
      cr->cookie = cookiebuf;
#ifdef DEBUG
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match: found %s", cookiebuf);
#endif
      return(0);
    }
  }
#ifdef DEBUG
/* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, cr->r, "TKT cookie_match: NOT found");*/
#endif
  return (1);
}


static char * 
find_our_cookie(request_rec *r, tkt_auth_dir_conf *tkt_cfg)
{
  cookie_res *cr = ap_palloc(r->pool, sizeof(cookie_res));
  
  cr->r = r;
  cr->cookie = NULL;
  cr->cookie_name = tkt_cfg->auth_cookie_name;
  
  ap_table_do(cookie_match, (void *) cr, r->headers_in, "Cookie", NULL);
  
  return (cr->cookie);
}

static char * 
find_our_url(request_rec *r, tkt_auth_dir_conf *tkt_cfg)
{
  const char *args = NULL;  /* url arguments string  */
  const char *key, *val;
  char *cookie = NULL;
  request_rec *main;

#ifdef DEBUG
  /*
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
    "TKT: no cookie found, checking url.");
  */
#endif

  /* Use main request args if subrequest */
  main = r->main == NULL ? r : r->main;
  if (main->args != NULL) {
    args = ap_pstrdup(r->pool, main->args); 
  }

  if (args != NULL) {
#ifdef DEBUG
    /*
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
      "TKT: looking for ticket in url: <%s>", args);
    */
#endif

    while (*args && (val = ap_getword(r->pool, &args, '&'))) {
      key = ap_getword(r->pool, &val, '=');

      if (strcmp(key,tkt_cfg->auth_cookie_name) == 0) {
	char *domain, *new_cookie;
	/* ap_unescape_url((char *)key); */
        /* ap_unescape_url((char *)val); */
#ifdef DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: Found url ticket: <%s>", val);
#endif

	/* Duplicate value into cookie for subsequent code */
	cookie = ap_pstrdup(r->pool, val);
	
	/* Default to current domain if auth domain not set */
        domain = tkt_cfg->auth_domain;
        if (!domain) {
          domain = strchr(ap_get_server_name(r), '.');
        }

	/* Set cookie now for subsequent auth */
	new_cookie = domain ?
	  ap_psprintf(r->pool,
            "%s=%s; path=/; domain=%s", 
            tkt_cfg->auth_cookie_name, val, domain) :
	  ap_psprintf(r->pool,
            "%s=%s; path=/", 
            tkt_cfg->auth_cookie_name, val);
#ifdef DEBUG
        /*
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, 
          "TKT: Setting new cookie from url: <%s>", new_cookie);
        */
#endif
	ap_table_add (r->headers_out, "Set-Cookie", new_cookie);

        /* Found ticket - ignore rest of arguments */
        break;
      }
    }
  } /*  end found url arguments */

  return cookie;
}


static ap_inline unsigned char *c2x(unsigned what, unsigned char *where)
{
  *where++ = '%';
  *where++ = c2x_table[what >> 4];
  *where++ = c2x_table[what & 0xf];
  return where;
}


static char * 
escape_str(pool *p, const char *segment)
{
  char *copy = ap_palloc(p, 3 * strlen(segment) + 1);
  const unsigned char *s = (const unsigned char *)segment;
  unsigned char *d = (unsigned char *)copy;
  unsigned c;
  
  while ((c = *s)) {
    if (TEST_CHAR(c, T_ESCAPE_PATH_SEGMENT) || c == '=' || c == '&') {
      d = c2x(c, d);
    }
    else {
      *d++ = c;
    }
    ++s;
  }
  *d = '\0';
  return copy;
}


static char * query_strip(request_rec *r, const char *strip)
{
  const char *args = NULL;  /* url arguments string  */
  const char *key, *val;
  request_rec *main;
  char *new_args = "";

  /* Use main request args if subrequest */
  main = r->main == NULL ? r : r->main;
  if (main->args == NULL || strip == NULL) 
    return NULL;

  args = ap_pstrdup(r->pool, main->args); 

  while (*args && (val = ap_getword(r->pool, &args, '&'))) {
    key = ap_getword(r->pool, &val, '=');

    /* Add to new_args only if key != strip */
    if (strlen(strip) != strlen(key) || strncmp(key,strip,strlen(strip)) != 0) 
      new_args = ap_psprintf(r->pool, "%s&%s=%s", new_args, key, val);
  }

  if (strlen(new_args) > 0)
    return new_args + 1;
}


static int 
redirect(request_rec *r, char *location, tkt_auth_dir_conf *tkt_cfg)
{
  /* Back cookies always use the current domain */
  char *domain = strchr(ap_get_server_name(r), '.');
  char *url;
  char *cookie;
  char sep;
  int sz;
  char *back_cookie_name = tkt_cfg->back_cookie_name;
  char *back_arg_name = tkt_cfg->back_arg_name;
  char *hostinfo, *back;

  /*
  tkt_auth_dir_conf *tkt_cfg = 
    ap_get_module_config(r->per_dir_config, &tkt_auth_module);
  
  */

  /* Get the scheme we use (http or https) and port */
  char *scheme = ap_http_method(r);
  int port = ap_get_server_port(r);

  /* Strip any auth_cookie_name arguments from the current args */
  char *query = query_strip(r, tkt_cfg->auth_cookie_name);
  if (query == NULL) {
    query = "";
  }
  else if (strlen(query) > 0) {
    query = ap_psprintf(r->pool, "?%s", query);
  }

  /* Set back URL */
  hostinfo = (port == ap_default_port_for_scheme(scheme)) ?
    ap_psprintf(r->pool, "%s", ap_get_server_name(r)) :
    ap_psprintf(r->pool, "%s:%d", ap_get_server_name(r), port);

  back = ap_psprintf(r->pool, 
    "%s%s%s%s%s", 
    scheme,
    escape_str ( r->pool, "://" ),
    hostinfo,
    escape_str ( r->pool, r->uri ),
    escape_str ( r->pool, query )
  );

#ifdef DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT back url: %s", back);
#endif
  
  /* Set back cookie if name is not null */
  if (back_cookie_name) {
    cookie = domain ?
      ap_psprintf(r->pool, 
        "%s=%s; path=/; domain=%s", 
        back_cookie_name, back, domain) :
      ap_psprintf(r->pool, 
        "%s=%s; path=/", 
        back_cookie_name, back);

    ap_table_setn(r->err_headers_out, "Set-Cookie", cookie);
    
    sz = 1 + strlen(location) + 1;
    url = ap_palloc(r->pool, sz);
    snprintf (url, sz, "%s", location);
  }

  /* If back_cookie_name not set, add a back url argument to location */
  else {
    sz = 1 + strlen(location) + 2 + strlen(back_arg_name) + strlen(back) + 1;
    url = ap_palloc(r->pool, sz);
    sep = (strchr(location, '?')) ? '&' : '?';
    snprintf (url, sz, "%s%c%s=%s", location, sep, back_arg_name, back);
  }

#ifdef DEBUG
  /* ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "TKT redirect: %s", url); */
#endif
  ap_table_setn(r->headers_out, "Location", url);
  
  return REDIRECT;
}


/*****************************************************************/

module MODULE_VAR_EXPORT tkt_auth_module =
{
	STANDARD_MODULE_STUFF,
	tkt_auth_init,		/* initializer */
	create_tkt_auth_config,	/* dir config creator */
	merge_tkt_auth_config,	/* dir merger --- default is to override */
	create_tkt_auth_serv_config,			/* server config */
	NULL,			/* merge server config */
	tkt_auth_cmds,	        /* command table */
	NULL,			/* handlers */
	NULL,			/* filename translation */
	authenticate_tkt_user,	/* check user_id */
	NULL,		        /* check auth */
	NULL,			/* check access */
	NULL,			/* type_checker */
	NULL,			/* fixups */
	NULL,			/* logger */
	NULL,			/* header parser */
	NULL,			/* chitkt_init */
	NULL,			/* chitkt_exit */
	NULL			/* post read-request */
};


/* vim:sw=2
 */
