#
# tag: utility perl routines for use with the Apache mod_auth_tkt module.
#

package TktUtil;

require Exporter;
use URI::Escape;
use Digest::MD5 qw(md5_hex);
use MIME::Base64;
use strict;

our @ISA = qw(Exporter);
our @EXPORT = qw(get_auth_ticket);
our @EXPORT_OK = qw(
    &encrypt
    &decrypt
    &get_secret
    &get_cookies
    &get_back_cookie
    &get_auth_ticket
);
our %EXPORT_TAGS = ( all => [@EXPORT_OK] );
our $VERSION = '0.06';

# our $SECRET_LABEL = 'MD5CryptKey';
our $SECRET_LABEL = 'TKTAuthSecret';
# our $SECRET_CONFIG_FILE = "/opt/web/conf/mod_auth_tkt.conf";
our $SECRET_CONFIG_FILE;
our $SECRET;

#
# Simple encryption routine using the shared secret with Blowfish
#
sub encrypt ($;$)
{
  my $plaintext = shift;
  my $secret = shift;
  $secret = get_secret() if ! $secret;
  require Crypt::CBC;
  my $cipher = Crypt::CBC->new($secret, 'Blowfish');
  my $ciphertext = $cipher->encrypt_hex($plaintext);
  return $ciphertext;
}

#
# Simple decryption routine using the shared secret with Blowfish
#
sub decrypt ($;$)
{
  my $ciphertext = shift;
  my $secret = shift;
  $secret = get_secret() if ! $secret;
  require Crypt::CBC;
  my $cipher = Crypt::CBC->new($secret, 'Blowfish');
  my $plaintext = $cipher->decrypt_hex($ciphertext);
  return $plaintext;
}

#
# Returns a shared secret from a config file
#
sub get_secret
{
    my %arg = @_;
    my $conf = $arg{conf} || $SECRET_CONFIG_FILE;
    my $ar = $arg{ar};                     # Apache request object (logging)
    die "[get_secret] no config file found" if ! $conf;

    # May be already cached
    return $SECRET if $SECRET;

    my $secret = '';
    if (!open CF, "<$conf") {
        $ar->warn("[get_secret] unable to open $conf") if $ar;
        die "[get_secret] unable to open $conf";
    }

    $/ = "\n";
    while (<CF>) {
        if (/^\s*$SECRET_LABEL\s+\"(.*)\"/) {
            $secret = $1;
            last;
        }
    }
    close CF;
    $ar->warn("[get_secret] $SECRET_LABEL not found in $conf")
        if (! $secret && $ar);

    # Cache
    $SECRET = $secret;

    return $secret;
}

my %cookies = ();
sub get_cookies
{
    my %arg = @_;
    my $ar = $arg{ar};
    unless ($ar) {
      require Apache;
      $ar = Apache->request;
    }
    my $cookie_header = $arg{cookies};
    $cookie_header ||= $ar->header_in('Cookie') if $ar;
    return undef unless $cookie_header;
    %cookies = map { split /=/, $_, 2; } split (/;\s*/, $cookie_header);
    return wantarray ? %cookies : \%cookies;
}

#
# Returns a back URL value from the back cookie (or a default).
#
sub get_back_cookie
{
    my %arg = @_;
    my $name = $arg{name} || 'TKTBack';
    my $def = $arg{default};

    my %cookies = get_cookies(%arg);
    return $def unless keys %cookies && $cookies{$name};

    my $back = uri_unescape($cookies{$name});
    if ($back && $back !~ m/^http/io) {
        my $http = 'http';
        $http = 'https' if $back =~ s/:443\b//o;
        $back = "$http://$back";
    }

    return $back;
}

#
# Generate a mod_auth_tkt cookie value for the given parameters.
#
sub new_ticket
{
    my %arg = @_;
    my $base64 = $arg{base64};
    my $secret = $arg{secret};
    my $uid = $arg{uid} || '';
    my $ip_addr = $arg{ip_addr} || '0.0.0.0';
    my $tokens = $arg{tokens} || '';
    # Remove whitespace around commas
    if ($tokens) {
        $tokens =~ s/,\s+/,/g;
        $tokens =~ s/\s+,/,/g;   
    }
    my $data = $arg{data} || '';
    my $ts = $arg{ts} || time;
    my $ar = $arg{ar};                     # Apache request object (logging)
    return undef if ! defined $uid;
    $secret ||= get_secret(ar => $ar);

    my @ip = split /\./, $ip_addr;
    my @ts = ( (($ts & 0xff000000) >> 24),
               (($ts & 0xff0000) >> 16),
               (($ts & 0xff00) >> 8),
               (($ts & 0xff)) );
    my $ipts = pack("C8", @ip, @ts);
    my $digest0 = md5_hex($ipts . $secret . $uid . $tokens . $data);
    my $digest = md5_hex($digest0 . $secret);

    $tokens .= '!' if $tokens;
    my $cookie =  sprintf "%s%08x%s!%s%s", $digest, $ts, $uid, $tokens, $data;
    # $ar->warn("[new_ticket:1] $cookie");
    $cookie = encode_base64($cookie, '') if $base64;
    # $ar->warn("[new_ticket:2] $cookie");

    return $cookie;
}

#
# Return the current uid from the TktAuth Cookie
#
sub get_cookie_uid
{
    my %arg = @_;
    my $name = $arg{name} || 'TKTAuth';
    my $ar = $arg{ar};
    my $cookies_in = $arg{cookies} || $ar->header_in('Cookie');

    my %cookies = map { split /=/, $_, 2; } split (/;\s*/, $cookies_in);
    return undef if ! exists $cookies{$name};
    my $uid = substr(decode_base64($cookies{$name}),40);
    $uid = substr($uid,0,index($uid,'!'));
    return $uid;
}

# 
# Returns a valid mod_auth_tkt cookie value (or cookie if $arg{cookie})
#   containing the given user details.
# This is a replacement routine for the tkt_cookie C-program in the old 
#   mod_auth_tkt.
#
sub get_auth_ticket
{
    my %arg = @_;
    my $uid = $arg{uid};
    my $ip_addr = $arg{ip_addr};
    my $tokens = $arg{tokens};
    my $data = $arg{data};
    my $ar = $arg{ar};
    my $errors = $arg{errors};
    my $log = $arg{log};
    my $cookie = $arg{cookie};
    my $name = $arg{name} || 'TKTAuth';
    my $domain = $arg{domain} || $ENV{HTTP_HOST};
    my $path = $arg{path} || '/';
    my $base64 = $arg{base64};
    my $secret = $arg{secret};
    my $secret_conf = $arg{secret_conf};
    my $ts = $arg{ts};                   # Debugging
    my $timeout = $arg{timeout};
    $log = 1 if ! defined $log && $ar;
    $base64 = 1 if ! defined $base64;      # Default to on
    die "[get_auth_ticket] no ip address parameter found" 
        if ! exists $arg{ip_addr};

    $uid = get_cookie_uid(name => $name, ar => $ar) if ! defined $uid;
    $uid = 0 if ! defined $uid;
    $ar->warn("[get_auth_ticket] timeout: refreshing cookie for uid $uid")
        if $log && $timeout;

    # Get MD5 shared secret
    if (! $secret) {
        $secret = get_secret(conf => $secret_conf, ar => $ar);
        if (! $secret) {
            push @$errors, "Error: MD5 shared secret not found."
                if ref $errors eq 'ARRAY';
            return undef;
        }
    }

    # Generate ticket cookie value
    my $ticket = new_ticket(base64 => $base64, uid => $uid, ts => $ts, 
        ip_addr => $ip_addr, data => $data, secret => $secret, 
        tokens => $tokens, ar => $ar);
    if (! $ticket) {
        push @$errors, "Error: problem generating cookie."
            if ref $errors eq 'ARRAY';
        return undef;
    }

    # Log if so requested
    if ($log) {
        my $out = "[get_auth_ticket] cookie: $ticket, domain: $domain, ip: $ip_addr, uid $uid, tokens $tokens, user_data: |$data|";
        if ($ar) {
            $ar->warn($out);
        }
        else {
            print STDERR $out, "\n";
        }
    }

    # Return the cookie value or the cookie
    if (! $cookie) {
        return $ticket;
    }
    else {
        return "$name=$ticket; domain=$domain; path=$path";
    }
}

1;

__END__

=head1 NAME

TktUtil - Utility routines for use with the Apache mod_auth_tkt module.


=head1 SYNOPSIS

    # Exports only get_auth_ticket by default
    use TktUtil qw(:all);

    # Either initialise secret OR tell TktUtil where to find it
    $TktUtil::SECRET = 'w b@5b15#664038f.f9d8U19b7e25 664eY9ad2%f';
    $TktUtil::SECRET_CONFIG_FILE = '/opt/web/conf/secret.conf';

    # Generate a new authentication ticket from which to make a cookie
    $ticket = get_auth_ticket(uid => $username, ip_addr => $ip_addr);

    # Or directly generate a new authentication cookie
    $cookie = get_auth_ticket(uid => $username, ip_addr => $ip_addr,
        cookie => 1);

    # Get the shared secret (for use with tkt_cookie executable)
    $secret = get_secret();

    # Convenience encryption and decryption routines (uses Blowfish)
    $user_data = encrypt($password);
    $password = decrypt($user_data);

    
=head1 DESCRIPTION

TktUtil is a set of routines for use with the mod_auth_tkt Apache module.
Its main purpose is to provide a means of generating valid mod_auth_tkt
tickets as an alternative to the standalone tkt_cookie executable. This is 
done by the get_auth_ticket routine, which is the only routine exported 
by default.

=head2 get_auth_ticket

get_auth_ticket generates mod_auth_tkt tickets for users. It uses a 
named-parameters calling style. Only the ip_addr parameter is 
actually required, although you'll typically want to provide a 
username/uid as well.

    $ticket = get_auth_ticket(uid => $username, ip_addr => $ip_addr);

uid defaults to 0 if not supplied; this is sometimes a useful setting for
providing guest or public access to certain areas.

By default, get_auth_ticket returns a base64-encoded value suitable for 
use as a cookie value - you still must generate a cookie yourself using,
for example, the CGI cookie() routine. Alternatively, you can get 
get_auth_ticket to produce a finished cookie itself using by setting the
'cookie' flag:

    $cookie = get_auth_ticket(uid => $username, ip_addr => $ip_addr,
        cookie => 1);

If you are using the TKTAuthCookieName and/or TKTAuthDomain directives
in your Apache config, then you will want to set the 'name' and 'domain'
parameters as well:

    $cookie = get_auth_ticket(uid => $username, ip_addr => $ip_addr,
        cookie => 1, name => 'perl_session', domain => '.foo.com');

The full list of get_auth_ticket parameters is as follows:

=over 4

=item uid

uid, username, or other identifier for this ticket. Defaults to 0 
if not set. Note that mod_auth_tkt does not require that this be a 
unique value.

=item ip_addr

IP address for this ticket (required), typically taken from the 
REMOTE_ADDR environment variable. If you are using the TKTAuthIgnoreIP
Apache directive, you should pass but false value for this parameter 
(e.g. undef).

=item data

Arbitrary user data to be stored for this ticket. mod_auth_tkt includes
this data in its MD5 hash checking so it should be able to be trusted.

=item errors

An arrayref used to return any error messages encountered.

=item log

Flag used to indicate whether to log the issue of this ticket (default 
0). If set, get_auth_ticket will log directly to the Apache log if the 
following parameter ($ar) is given; otherwise it prints to STDERR 
(usually the same effect).

=item ar

The Apache request record - this is available when running under mod_perl
and is only used for logging.

=item cookie

Flag used to indicate whether to produce a full cookie (instead of just
a cookie value). Defaults to 0.

=item name

Cookie name to use if producing a full cookie. Should match the
TKTAuthCookieName directive, if you're using it. Defaults to TKTAuth.

=item domain

Cookie domain to use if producing a full cookie. Should match the
TKTAuthDomain directive, if you're using it. Defaults to the value 
of the HTTP_HOST environment variable.

=item path

Cookie path to use if producing a full cookie. Defaults to '/'.

=item base64

Flag used to indicate whether to base64 encode the cookie value (default 1).

=back

=head1 BUGS and LIMITATIONS

It would be nice to be able to mine the mod_auth_tkt Apache 
configuration directives for default settings for get_auth_ticket,
instead of forcing the user to pass them again as parameters.


=head1 AUTHOR

Gavin Carr (gavin@openfusion.com.au)


=head1 COPYRIGHT

Copyright 2001-2002 Open Fusion Pty Ltd. All Rights Reserved.

This program is free software. You may copy or redistribute it under the 
same terms as perl itself.

=cut
