#!/usr/bin/perl -w
#
# mod_auth_tkt login script - runs both as a vanilla CGI and under mod_perl
#   via Apache::Registry.
#
# This script can run in a few different modes, depending on how it is 
#   named. Copy the script to a cgi-bin area, and create appropriately 
#   named symlinks to access the different behaviours. 
# Modes:
#   - login mode (default): request a username and password and test via
#     $validate_sub - if successful, issue an auth ticket and redirect to 
#     the back location
#   - guest mode ('guest.cgi'): automatically issues an auth ticket a 
#     special username (as defined in $guest_sub, default 'guest'), and 
#     redirect to the back location
#   - autologin mode ('autologin.cgi'): [typically used to allow tickets 
#     across multiple domains] if no valid auth ticket exists, redirect
#     to the login (or guest) version; otherwise automatically redirect 
#     to the back location passing the current auth ticket as a GET 
#     argument. mod_auth_tkt (>= 1.3.8) will turn this new ticket into 
#     an auth cookie for the new domain if none already exists.
#

# use lib '/path/to/TktUtil';
use TktUtil qw(get_auth_ticket encrypt);
# use Apache::Htpasswd;
use File::Basename;
use CGI qw(:standard);
use URI::Escape;
use URI;
use strict;

# ------------------------------------------------------------------------
# Configure this section to taste

# Cookie names
my $AUTH_COOKIE_NAME = 'auth_tkt';
my $PROBE_COOKIE_NAME = 'auth_probe';
my $BACK_COOKIE_NAME = 'back';
my $BACK_ARG_NAME = 'back';
# Mode to fallback to if autologin fails ('login' or 'guest')
my $AUTOLOGIN_FALLBACK_MODE = 'login';

# For login mode (if used), setup username/password validation
#   (modify or point $validate_sub somewhere appropriate).
sub validate
{
  my ($username, $password) = @_;
  # Validate using Apache::Htpasswd and a standard htpasswd file
# my $ht = Apache::Htpasswd->new({ 
#   passwdFile => '/etc/httpd/conf/htpasswd', ReadOnly => 1 });
# return $ht->htCheckPassword($username, $password);
  # Noop - always fails
  return 0;
}
my $validate_sub = \&validate;

# For guest mode (if used), setup guest username
sub guest_user
{
  return 'guest';
}
my $guest_sub = \&guest_user;

# ------------------------------------------------------------------------
# Setup some initial parameters
my $debug = 0;
my $q = CGI->new;
$TktUtil::SECRET_CONFIG_FILE = $ENV{MOD_AUTH_TKT_CONF};
my $AUTH_DOMAIN = $ENV{AUTH_DOMAIN};
unless ($AUTH_DOMAIN) {
  # If no AUTH_DOMAIN set, default to server wildcard domain (.foo.com)
  $AUTH_DOMAIN = $ENV{SERVER_NAME};
  $AUTH_DOMAIN =~ s/^[^\.]*// if $AUTH_DOMAIN;
}
my $ticket = $q->cookie($AUTH_COOKIE_NAME);
my $probe = $q->cookie($PROBE_COOKIE_NAME);
my $back = $q->cookie($BACK_COOKIE_NAME);
my $have_cookies = $ticket || $probe || $back || '';
$back ||= $q->param($BACK_ARG_NAME);
if ($back && $back =~ m!^/!) {
  my $hostname = $ENV{SERVER_NAME};
  my $port = $ENV{SERVER_PORT};
  $hostname .= ':' . $port if $port && $port != 80 && $port != 443;
  $back = sprintf "http%s://%s%s", ($port == 443 ? 's' : ''), $hostname, $back;
} elsif ($back && $back !~ m/^http/i) {
  $back = 'http://' . $back;
}
$back = uri_unescape($back);
my $back_esc = uri_escape($back);

my ($fatal, @errors);
my ($mode, $location, $suffix) = fileparse($ENV{SCRIPT_NAME}, '\.cgi', '\.pl');
$mode = 'login' unless $mode eq 'guest' || $mode eq 'autologin';
my $self_redirect = $q->param('redirect') || '';
my $username = lc($q->param('username'));
my $password = $q->param('password');
my $timeout = $q->param('timeout');
my $unauth = $q->param('unauth');
my $ip_addr = $ENV{REMOTE_ADDR};
my $redirected = 0;
my $cookie_fmt = $AUTH_DOMAIN ? "%s=%s; domain=%s; path=/" : "%s=%s; path=/";

# ------------------------------------------------------------------------
# Set the auth cookie and redirect to $back
my $set_cookie_redirect = sub {
  my ($tkt, $back) = @_;

  if ($back) {
    my $b = URI->new($back);
    # If $back domain doesn't match $AUTH_DOMAIN, pass ticket via back GET param
    my $domain = $AUTH_DOMAIN || $ENV{SERVER_NAME};
    if ($b->host !~ m/\b$domain$/i) {
      $back .= $b->query ? '&' : '?';
      $back .= $AUTH_COOKIE_NAME . '=' . $tkt;
    }
    # Set (local) cookie, and redirect to $back
    print $q->header(
      -cookie => sprintf($cookie_fmt, $AUTH_COOKIE_NAME, $tkt, $AUTH_DOMAIN),
      -location => $back,
    );

    return 1;
  }

  # If no $back, just set the auth cookie and hope for the best
  else {
    print $q->header(
      -cookie => sprintf($cookie_fmt, $AUTH_COOKIE_NAME, $tkt, $AUTH_DOMAIN)
    );
  }
};

# ------------------------------------------------------------------------
# Actual processing

# If no cookies found, first check whether cookies are supported
if (! $have_cookies) {
  # If this is a self redirect warn the user about cookie support
  if ($self_redirect) {
    $fatal = "Your browser does not appear to support cookies or has cookie support disabled.<br />\nThis site requires cookies - please turn cookie support on or try again using a different browser.";
  }
  # If no cookies and not a redirect, redirect to self to test cookies
  else {
    my $extra = '';
    $extra .= 'timeout=1' if $timeout;
    $extra .= 'unauth=1' if $unauth;
    $extra = "&$extra" if $extra;
    print $q->header(
      -cookie => sprintf($cookie_fmt, $PROBE_COOKIE_NAME, 1, $AUTH_DOMAIN),
      -location => sprintf("%s%s%s?redirect=1&%s=%s%s",
        $location, $mode, $suffix, $BACK_COOKIE_NAME, $back_esc, $extra),
    );
    $redirected = 1;
  }
}

elsif ($mode eq 'autologin') {
  # If we have a ticket, redirect to $back, including ticket as GET param
  if ($ticket && $back && ! $timeout) {
    my $b = URI->new($back);
    $back .= $b->query ? '&' : '?';
    $back .= $AUTH_COOKIE_NAME . '=' . $ticket;
    print $q->redirect($back);
    $redirected = 1;
  }
  # Can't autologin - change mode to either guest or login
  else {
    $mode = $AUTOLOGIN_FALLBACK_MODE;
  }
}

unless ($fatal || $redirected) {
  if ($mode eq 'login') {
    if ($username && $validate_sub->($username, $password)) {
#     my $user_data = join(':', encrypt($password), time(), $ip_addr);
      my $user_data = join(':', time(), $ip_addr);    # Optional
      my $tkt = get_auth_ticket(uid => $username, data => $user_data,
	ip_addr => $ENV{REMOTE_ADDR}, errors => \@errors, cookie => 0); 
      if (! @errors) {
        $redirected = $set_cookie_redirect->($tkt, $back);
        $fatal = "Login successful.";
      }
    }
    elsif ($username) {
      push @errors, "Invalid username or password.";
    }
  }

  elsif ($mode eq 'guest') {
    # Generate a guest cookie and redirect to $back
    my $tkt = get_auth_ticket(uid => $guest_sub->(), 
      ip_addr => $ENV{REMOTE_ADDR}, errors => \@errors, cookie => 0); 
    if (! @errors) {
      $redirected = $set_cookie_redirect->($tkt, $back);
      $fatal = "No back link found.";
    }
  }
}

unless ($redirected) {
  # If here, either some kind of error or a login page
  if ($fatal) {
    print $q->header,
      $q->start_html(
        -title => "\u$mode Page",
      );
  }
  else {
    push @errors, qq(Your session has timed out.) if $timeout;
    push @errors, qq(You are not authorised to access this area.) if $unauth;
    print $q->header,
      $q->start_html(
        -title => "\u$mode Page",
        -onLoad => "getFocus()",
        -script => qq(
function getFocus() {
  document.forms[0].elements[0].focus();
  document.forms[0].elements[0].select();
}));
  }
  print <<EOD;
<div align="center">
<h1>\u$mode Page</h1>
EOD

  if ($debug) {
    print <<EOD;
<pre>
SERVER_NAME: $ENV{SERVER_NAME}
domain: $AUTH_DOMAIN
mode: $mode
suffix: $suffix
have_cookies: $have_cookies
</pre>
EOD
  }

  if ($fatal) {
    print qq(<p class="error">$fatal</p>\n);
  }

  else {
    print qq(<p class="error">\n), join(qq(<br />\n), @errors), "</p>\n"
      if @errors;
    print <<EOD;
<form name="login" method="post" action="$mode$suffix">
<table border="0" cellpadding="5">
<tr><th>Username:</th><td><input type="text" name="username" /></td></tr>
<tr><th>Password:</th><td><input type="password" name="password" /></td></tr>
<tr><td colspan="2" align="center"> 
<input type="submit" value="Login">
</td></tr>
</table>
EOD
    print qq(<input type="hidden" name="back" value="$back_esc">\n) if $back_esc;
    print qq(</form>\n);
}

  print qq(<p><a href="$back">Previous Page</a></p>\n) if $back;
  print <<EOD;
</div>
</body>
</html>
EOD
}

# tag: mod_auth_tkt example login CGI script
# vim:sw=2:sm:cin

